/********************************************************************************************
 *     LEGAL DISCLAIMER
 *
 *     (Header of MediaTek Software/Firmware Release or Documentation)
 *
 *     BY OPENING OR USING THIS FILE, BUYER HEREBY UNEQUIVOCALLY ACKNOWLEDGES AND AGREES
 *     THAT THE SOFTWARE/FIRMWARE AND ITS DOCUMENTATIONS ("MEDIATEK SOFTWARE") RECEIVED
 *     FROM MEDIATEK AND/OR ITS REPRESENTATIVES ARE PROVIDED TO BUYER ON AN "AS-IS" BASIS
 *     ONLY. MEDIATEK EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES, EXPRESS OR IMPLIED,
 *     INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
 *     A PARTICULAR PURPOSE OR NONINFRINGEMENT. NEITHER DOES MEDIATEK PROVIDE ANY WARRANTY
 *     WHATSOEVER WITH RESPECT TO THE SOFTWARE OF ANY THIRD PARTY WHICH MAY BE USED BY,
 *     INCORPORATED IN, OR SUPPLIED WITH THE MEDIATEK SOFTWARE, AND BUYER AGREES TO LOOK
 *     ONLY TO SUCH THIRD PARTY FOR ANY WARRANTY CLAIM RELATING THERETO. MEDIATEK SHALL ALSO
 *     NOT BE RESPONSIBLE FOR ANY MEDIATEK SOFTWARE RELEASES MADE TO BUYER'S SPECIFICATION
 *     OR TO CONFORM TO A PARTICULAR STANDARD OR OPEN FORUM.
 *
 *     BUYER'S SOLE AND EXCLUSIVE REMEDY AND MEDIATEK'S ENTIRE AND CUMULATIVE LIABILITY WITH
 *     RESPECT TO THE MEDIATEK SOFTWARE RELEASED HEREUNDER WILL BE, AT MEDIATEK'S OPTION,
 *     TO REVISE OR REPLACE THE MEDIATEK SOFTWARE AT ISSUE, OR REFUND ANY SOFTWARE LICENSE
 *     FEES OR SERVICE CHARGE PAID BY BUYER TO MEDIATEK FOR SUCH MEDIATEK SOFTWARE AT ISSUE.
 *
 *     THE TRANSACTION CONTEMPLATED HEREUNDER SHALL BE CONSTRUED IN ACCORDANCE WITH THE LAWS
 *     OF THE STATE OF CALIFORNIA, USA, EXCLUDING ITS CONFLICT OF LAWS PRINCIPLES.
 ************************************************************************************************/
/*
*
* \par Description
*   The function bodies of mt85 surface manager
*
*/

#include <fusion/shmalloc.h>

#include <directfb.h>
#include <directfb_util.h>

#include <core/core.h>

#include <core/gfxcard.h>
#include <core/surface.h>
#include <core/surface_buffer.h>

#include <direct/debug.h>
#include <direct/messages.h>
#include <direct/util.h>

#include <gfx/convert.h>

#include <misc/conf.h>

#include <sys/ioctl.h>
#include "mt85.h"

#include "mt85_surfacemanager.h"



D_DEBUG_DOMAIN( SurfMan, "SurfaceManager", "DirectFB Surface Manager" );


static Chunk *split_chunk ( SurfaceManager *manager,
                            Chunk          *chunk,
                            int             length );

static Chunk *free_chunk  ( SurfaceManager *manager,
                            Chunk          *chunk );

static Chunk *occupy_chunk( SurfaceManager     *manager,
                            Chunk              *chunk,
                            CoreSurfaceAllocation *allocation,
                            int                 length,
                            int                 pitch );


DFBResult
dfb_surfacemanager_create( CoreDFB         *core,
                           unsigned int     length,
                           SurfaceManager **ret_manager )
{


    FusionSHMPoolShared *pool;
    SurfaceManager      *manager;
    Chunk               *chunk;
    int                  i = 0;

    D_DEBUG_AT( SurfMan, "%s( %p, %d )\n", __FUNCTION__, core, length );

    D_ASSERT( core != NULL );
    D_ASSERT( ret_manager != NULL );

    pool = dfb_core_shmpool( core );

    manager = SHCALLOC( pool, 1, sizeof(SurfaceManager) );
    if (!manager)
        return D_OOSHM();

    for (i = 0; i < mt85_multi_pool.num; i++)
    {
        chunk = SHCALLOC( pool, 1, sizeof(Chunk) );
        if (!chunk)
        {
            D_OOSHM();
            SHFREE( pool, manager );
            return DFB_NOSHAREDMEMORY;
        }

        manager->m_chunks[i]  = chunk;
        manager->m_offset[i]  = 0;
        manager->m_length[i]  = mt85_multi_pool.size[i];
        manager->m_avail[i]   = manager->m_length[i] - manager->m_offset[i];

        chunk->offset    = manager->m_offset[i];
        chunk->length    = manager->m_avail[i];
        chunk->poolno    = i;

        D_MAGIC_SET( chunk, Chunk );
    }

    manager->shmpool = pool;

    D_MAGIC_SET( manager, SurfaceManager );

    D_DEBUG_AT( SurfMan, "  -> %p\n", manager );

    *ret_manager = manager;

    return DFB_OK;

}

void
dfb_surfacemanager_destroy( SurfaceManager *manager )
{
    Chunk *chunk;
    void  *next;
    int i = 0;

    D_DEBUG_AT( SurfMan, "%s( %p )\n", __FUNCTION__, manager );

    D_MAGIC_ASSERT( manager, SurfaceManager );

    /* Deallocate all video chunks. */
    for (i = 0; i < mt85_multi_pool.num; i++)
    {
        chunk = manager->m_chunks[i];
        while (chunk)
        {
            next = chunk->next;

            D_MAGIC_CLEAR( chunk );

            SHFREE( manager->shmpool, chunk );

            chunk = next;
        }
    }

    D_MAGIC_CLEAR( manager );

    /* Deallocate manager struct. */
    SHFREE( manager->shmpool, manager );
}

DFBResult search_pool( SurfaceManager *manager, Chunk **c, CoreSurfaceBuffer *buffer, Chunk **best_free, Chunk **ret_chunk, int length, int poolno )
{
    /* FIXME_SC_2  Workaround creation happening before graphics driver initialization. */
    if (!(*c)->next)
    {
        int length = mt85_multi_pool.size[poolno];

        if ((*c)->length != length - manager->m_offset[poolno])
        {
            D_WARN( "workaround" );

            manager->m_length[poolno] = length;
            manager->m_avail[poolno]  = length - manager->m_offset[poolno];

            (*c)->length = length - manager->m_offset[poolno];
        }

    }

    while ((*c))
    {
        D_MAGIC_ASSERT( (*c), Chunk );

        if (!(*c)->buffer && (*c)->length >= length)
        {
            /* NULL means check only. */
            if (!ret_chunk)
                return DFB_FAILURE;

            /* found a nice place to chill */
            if (!(*best_free)  ||  (*best_free)->length > (*c)->length)
                /* first found or better one? */
                (*best_free) = (*c);

            if ((*c)->length == length)
                break;
        }

        (*c) = (*c)->next;
    }

    return DFB_OK;
}

/** public functions NOT locking the surfacemanger theirself,
    to be called between lock/unlock of surfacemanager **/

DFBResult dfb_surfacemanager_allocate( CoreDFB            *core,
                                       SurfaceManager     *manager,
                                       CoreSurfaceBuffer  *buffer,
                                       CoreSurfaceAllocation  *allocation,
                                       Chunk             **ret_chunk )
{

    int pitch;
    int length;
    Chunk *c;
    CoreGraphicsDevice *device;

    int poolno;
    int i,j;

    Chunk *best_free     = NULL;

    DFBResult ret = DFB_TEMPUNAVAIL;

    D_MAGIC_ASSERT( manager, SurfaceManager );
    D_MAGIC_ASSERT( buffer, CoreSurfaceBuffer );
    D_MAGIC_ASSERT( buffer->surface, CoreSurface );

    if (ret_chunk)
        D_MAGIC_ASSERT( allocation, CoreSurfaceAllocation );
    else
        D_ASSUME( allocation == NULL );

    D_DEBUG_AT( SurfMan, "%s( %p ) <- %dx%d %s\n", __FUNCTION__, buffer,
                buffer->surface->config.size.w, buffer->surface->config.size.h,
                dfb_pixelformat_name( buffer->surface->config.format ) );

    if (manager->suspended)
        return DFB_SUSPENDED;

    /* FIXME: Only one global device at the moment. */
    device = dfb_core_get_part( core, DFCP_GRAPHICS );
    D_ASSERT( device != NULL );

    dfb_gfxcard_calc_buffer_size( device, buffer, &pitch, &length );

    DFB_MEM_LOG("[DirectFB][Memory][SurfMan] %s:%d ->allocate pitch %d,length %d\n",__FUNCTION__,__LINE__,pitch, length );

    D_DEBUG_AT( SurfMan, "  -> pitch %d, length %d\n", pitch, length );
    if (!(mt85_multi_pool.iommu_support_flag))
    {
        if (dfb_config->allocate_mem_reverse_order)
        {
            DFB_MEM_LOG("[DirectFB][Memory][SurfMan]No iommu mode and allocate by reverse order\n");
            for (i = mt85_multi_pool.num - 1; i >= 0 ; i--)
            {
                if (manager->m_avail[i] >= length)
                {
                    ret = DFB_OK;
                    break;
                }
            }
            if (DFB_TEMPUNAVAIL == ret)
                return DFB_TEMPUNAVAIL;

            /* examine chunks */
            for (i = mt85_multi_pool.num - 1; i >= 0 ; i--)
            {
                c = manager->m_chunks[i];
                D_MAGIC_ASSERT( c, Chunk );

                ret = search_pool( manager, &c, buffer, &best_free, ret_chunk, length, i );
                if (DFB_FAILURE == ret)
                    return DFB_OK;
                if (best_free)
                {
                    poolno = i;
                    break;
                }
            }
        }
        else
        {
            DFB_MEM_LOG("[DirectFB][Memory][SurfMan]No iommu mode and allocate by inverse order\n");
            for (i = 0; i < mt85_multi_pool.num; i++)
            {
                if (manager->m_avail[i] >= length)
                {
                    ret = DFB_OK;
                    break;
                }
            }
            if (DFB_TEMPUNAVAIL == ret)
                return DFB_TEMPUNAVAIL;

            /* examine chunks */
            for (i = 0; i < mt85_multi_pool.num; i++)
            {
                c = manager->m_chunks[i];
                D_MAGIC_ASSERT( c, Chunk );

                ret = search_pool( manager, &c, buffer, &best_free, ret_chunk, length, i );
                if (DFB_FAILURE == ret)
                    return DFB_OK;
                if (best_free)
                {
                    poolno = i;
                    break;
                }
            }

        }
    }
    else
    {
        DFB_MEM_LOG("[DirectFB][Memory][SurfMan]IOMMU MODE\n");
        if ((buffer->surface->config.caps & DSCAPS_PRIMARY) && (buffer->surface->config.caps & DSCAPS_VIDEOONLY))
        {
            for (i = 0; (i < mt85_multi_pool.num) && (mt85_multi_pool.type[i] != FBMT_VALLOC); i++)
            {
                if (manager->m_avail[i] >= length)
                {
                    ret = DFB_OK;
                    break;
                }
            }
            if (DFB_TEMPUNAVAIL == ret)
                return DFB_TEMPUNAVAIL;

            /* examine chunks */
            for (i = 0; (i < mt85_multi_pool.num) && (mt85_multi_pool.type[i] != FBMT_VALLOC); i++)
            {
                c = manager->m_chunks[i];
                D_MAGIC_ASSERT( c, Chunk );

                /*printf("......Searching Pool %d......\n", i);*/
                ret = search_pool( manager, &c, buffer, &best_free, ret_chunk, length, i );
                if (DFB_FAILURE == ret)
                    return DFB_OK;
                if (best_free)
                {
                    /*printf("......Found\n");*/
                    poolno = i;
                    break;
                }
            }
        }
        else //alloc from last block for normal surface
        {
            for (i = mt85_multi_pool.num - 1; i >= 0 ; i--)
            {
                if (manager->m_avail[i] >= length)
                {
                    ret = DFB_OK;
                    break;
                }
            }
            if (DFB_TEMPUNAVAIL == ret)
                return DFB_TEMPUNAVAIL;

            /* examine chunks */
            for (i = mt85_multi_pool.num - 1; i >= 0 ; i--)
            {
                c = manager->m_chunks[i];
                D_MAGIC_ASSERT( c, Chunk );

                /*printf("......Searching Pool %d......\n", i);*/
                ret = search_pool( manager, &c, buffer, &best_free, ret_chunk, length, i );
                if (DFB_FAILURE == ret)
                    return DFB_OK;
                if (best_free)
                {
                    /*printf("......Found\n");*/
                    poolno = i;
                    break;
                }
            }
        }
    }

    /* if we found a place */
    if (best_free)
    {
        DFB_MEM_LOG("[DirectFB][Memory][SurfMan] %s:%d -> Found free Size = %d\n",__FUNCTION__,__LINE__,best_free->length);

        /* NULL means check only. */
        if (ret_chunk)
        {
            *ret_chunk = occupy_chunk( manager, best_free, allocation, length, pitch );
            (*ret_chunk)->poolno = poolno;
            //clean the memory if  it is used in window & layer surface
            if (buffer->surface->type  & (CSTF_LAYER|CSTF_WINDOW) )
            {
                memset((void*)(mt85_multi_pool.addr[(*ret_chunk)->poolno]+(*ret_chunk)->offset), 0,(*ret_chunk)->length);
            }
            for (j = 0; j < mt85_multi_pool.num; j++)
            {
                DFB_MEM_LOG("[DirectFB][Memory][SurfMan] PooL Num = %d ,Pool Total Size = %d,Pool Available Size = %d\n",j,manager->m_length[j], manager->m_avail[j]);
            }
            DFB_MEM_LOG("[DirectFB][Memory][SurfMan] %s:%d Current Allocated Size = %d In Pool %d\n",__FUNCTION__,__LINE__,length,(*ret_chunk)->poolno);
        }

        return DFB_OK;
    }

    /* no luck */
    return DFB_NOVIDEOMEMORY;


}

DFBResult dfb_surfacemanager_displace( CoreDFB           *core,
                                       SurfaceManager    *manager,
                                       CoreSurfaceBuffer *buffer )
{
    int                    length;
    Chunk                 *multi_start = NULL;
    int                    multi_size  = 0;
    int                    multi_tsize = 0;
    int                    multi_count = 0;
    Chunk                 *bestm_start = NULL;
    int                    bestm_count = 0;
    int                    bestm_size  = 0;
    int                    min_toleration;
    Chunk                 *chunk;
    CoreGraphicsDevice    *device;
    CoreSurfaceAllocation *smallest = NULL;

    D_MAGIC_ASSERT( manager, SurfaceManager );
    D_MAGIC_ASSERT( buffer, CoreSurfaceBuffer );
    D_MAGIC_ASSERT( buffer->surface, CoreSurface );

    D_DEBUG_AT( SurfMan, "%s( %p ) <- %dx%d %s\n", __FUNCTION__, buffer,
                buffer->surface->config.size.w, buffer->surface->config.size.h,
                dfb_pixelformat_name( buffer->surface->config.format ) );

    /* FIXME: Only one global device at the moment. */
    device = dfb_core_get_part( core, DFCP_GRAPHICS );
    D_ASSERT( device != NULL );

    dfb_gfxcard_calc_buffer_size( dfb_core_get_part( core, DFCP_GRAPHICS ), buffer, NULL, &length );

    min_toleration = manager->m_min_toleration[mt85_multi_pool.num - 1]/8 + 2;

    D_DEBUG_AT( SurfMan, "  -> %7d required, min toleration %d\n", length, min_toleration );

    chunk = manager->m_chunks[mt85_multi_pool.num - 1];
    while (chunk)
    {
        CoreSurfaceAllocation *allocation;

        D_MAGIC_ASSERT( chunk, Chunk );

        allocation = chunk->allocation;
        if (allocation)
        {
            CoreSurfaceBuffer *other;
            int                size;

            D_MAGIC_ASSERT( allocation, CoreSurfaceAllocation );
            D_ASSERT( chunk->buffer == allocation->buffer );
            D_ASSERT( chunk->length >= allocation->size );

            other = allocation->buffer;
            D_MAGIC_ASSERT( other, CoreSurfaceBuffer );

            if (other->locked)
            {
                D_DEBUG_AT( SurfMan, "  ++ %7d locked %dx\n", allocation->size, other->locked );
                goto next_reset;
            }

            if (other->policy > buffer->policy)
            {
                D_DEBUG_AT( SurfMan, "  ++ %7d policy %d > %d\n", allocation->size, other->policy, buffer->policy );
                goto next_reset;
            }

            if (other->policy == CSP_VIDEOONLY)
            {
                D_DEBUG_AT( SurfMan, "  ++ %7d policy videoonly\n", allocation->size );
                goto next_reset;
            }

            chunk->tolerations++;
            if (chunk->tolerations > 0xff)
                chunk->tolerations = 0xff;

            if (other->policy == buffer->policy && chunk->tolerations < min_toleration)
            {
                D_DEBUG_AT( SurfMan, "  ++ %7d tolerations %d/%d\n",
                            allocation->size, chunk->tolerations, min_toleration );
                goto next_reset;
            }

            size = allocation->size;

            if (chunk->prev && !chunk->prev->allocation)
                size += chunk->prev->length;

            if (chunk->next && !chunk->next->allocation)
                size += chunk->next->length;

            if (size >= length)
            {
                if (!smallest || smallest->size > allocation->size)
                {
                    D_DEBUG_AT( SurfMan, "  => %7d [%d] < %d, tolerations %d\n",
                                allocation->size, size, smallest ? smallest->size : 0, chunk->tolerations );

                    smallest = allocation;
                }
                else
                    D_DEBUG_AT( SurfMan, "  -> %7d [%d] > %d\n", allocation->size, size, smallest->size );
            }
            else
                D_DEBUG_AT( SurfMan, "  -> %7d [%d]\n", allocation->size, size );
        }
        else
            D_DEBUG_AT( SurfMan, "  -  %7d free\n", chunk->length );


        if (!smallest)
        {
            if (!multi_start)
            {
                multi_start = chunk;
                multi_tsize = chunk->length;
                multi_size  = chunk->allocation ? chunk->length : 0;
                multi_count = chunk->allocation ? 1 : 0;
            }
            else
            {
                multi_tsize += chunk->length;
                multi_size  += chunk->allocation ? chunk->length : 0;
                multi_count += chunk->allocation ? 1 : 0;

                while (multi_tsize >= length && multi_count > 1)
                {
                    if (!bestm_start || bestm_size > multi_size * multi_count / bestm_count)
                    {
                        D_DEBUG_AT( SurfMan, "                =====> %7d, %7d %2d used [%7d %2d]\n",
                                    multi_tsize, multi_size, multi_count, bestm_size, bestm_count );

                        bestm_size  = multi_size;
                        bestm_start = multi_start;
                        bestm_count = multi_count;
                    }
                    else
                        D_DEBUG_AT( SurfMan, "                -----> %7d, %7d %2d used\n",
                                    multi_tsize, multi_size, multi_count );

                    if (multi_count <= 2)
                        break;

                    if (!multi_start->allocation)
                    {
                        multi_tsize -= multi_start->length;
                        multi_start  = multi_start->next;
                    }

                    D_ASSUME( multi_start->allocation != NULL );

                    multi_tsize -= multi_start->length;
                    multi_size  -= multi_start->allocation ? multi_start->length : 0;
                    multi_count -= multi_start->allocation ? 1 : 0;
                    multi_start  = multi_start->next;
                }
            }
        }

        chunk = chunk->next;

        continue;


next_reset:
        multi_start = NULL;

        chunk = chunk->next;
    }

    if (smallest)
    {
        D_MAGIC_ASSERT( smallest, CoreSurfaceAllocation );
        D_MAGIC_ASSERT( smallest->buffer, CoreSurfaceBuffer );

        smallest->flags |= CSALF_MUCKOUT;

        D_DEBUG_AT( SurfMan, "  -> offset %lu, size %d\n", smallest->offset, smallest->size );

        return DFB_OK;
    }

    if (bestm_start)
    {
        chunk = bestm_start;

        while (bestm_count)
        {
            CoreSurfaceAllocation *allocation = chunk->allocation;

            if (allocation)
            {
                D_MAGIC_ASSERT( allocation, CoreSurfaceAllocation );
                D_MAGIC_ASSERT( allocation->buffer, CoreSurfaceBuffer );

                allocation->flags |= CSALF_MUCKOUT;

                bestm_count--;
            }

            D_DEBUG_AT( SurfMan, "  ---> offset %d, length %d\n", chunk->offset, chunk->length );

            chunk = chunk->next;
        }

        return DFB_OK;
    }

    return DFB_NOVIDEOMEMORY;
}

DFBResult dfb_surfacemanager_deallocate( SurfaceManager *manager,
        Chunk          *chunk )
{
    CoreSurfaceBuffer *buffer;
    int j;

    D_MAGIC_ASSERT( manager, SurfaceManager );
    D_MAGIC_ASSERT( chunk, Chunk );

    buffer = chunk->buffer;
    D_MAGIC_ASSERT( buffer, CoreSurfaceBuffer );
    D_MAGIC_ASSERT( buffer->surface, CoreSurface );

    DFB_MEM_LOG("[DirectFB][Memory][SurfMan] %s:%d %s( %p ) <- %dx%d %s\n",__FUNCTION__,__LINE__, buffer,
                buffer->surface->config.size.w, buffer->surface->config.size.h,
                dfb_pixelformat_name( buffer->surface->config.format ) );

    DFB_MEM_LOG("[DirectFB][Memory][SurfMan] %s:%d Current Free Size = %dK,In Pool %d\n",__FUNCTION__,__LINE__,chunk->length,chunk->poolno);
    free_chunk( manager, chunk );

    for (j = 0; j < mt85_multi_pool.num; j++)
    {
        DFB_MEM_LOG("[DirectFB][Memory][SurfMan] PooL Num = %d ,Pool Total Size = %d,Pool Available Size = %d\n",j,manager->m_length[j], manager->m_avail[j]);
    }
    return DFB_OK;
}

/** internal functions NOT locking the surfacemanager **/

static Chunk *
split_chunk( SurfaceManager *manager, Chunk *c, int length )
{
    Chunk *newchunk;

    D_MAGIC_ASSERT( c, Chunk );

    if (c->length == length)          /* does not need be splitted */
        return c;

    newchunk = (Chunk*) SHCALLOC( manager->shmpool, 1, sizeof(Chunk) );
    if (!newchunk)
    {
        D_OOSHM();
        return NULL;
    }

    /* calculate offsets and lengths of resulting chunks */
    newchunk->offset = c->offset + c->length - length;
    newchunk->length = length;
    c->length -= newchunk->length;

    /* insert newchunk after chunk c */
    newchunk->prev = c;
    newchunk->next = c->next;
    if (c->next)
        c->next->prev = newchunk;
    c->next = newchunk;

    D_MAGIC_SET( newchunk, Chunk );

    return newchunk;
}

static Chunk *
free_chunk( SurfaceManager *manager, Chunk *chunk )
{
    D_MAGIC_ASSERT( manager, SurfaceManager );
    D_MAGIC_ASSERT( chunk, Chunk );

    if (!chunk->buffer)
    {
        D_BUG( "freeing free chunk" );
        return chunk;
    }
    else
    {
        D_DEBUG_AT( SurfMan, "Deallocating %d bytes at offset %d.\n", chunk->length, chunk->offset );
    }

    if (chunk->buffer->policy == CSP_VIDEOONLY)
        manager->m_avail[chunk->poolno] += chunk->length;

    chunk->allocation = NULL;
    chunk->buffer = NULL;

    manager->m_min_toleration[chunk->poolno]--;

    if (chunk->prev  &&  !chunk->prev->buffer)
    {
        Chunk *prev = chunk->prev;

        //D_DEBUG_AT( SurfMan, "  -> merging with previous chunk at %d\n", prev->offset );

        prev->length += chunk->length;

        prev->next = chunk->next;
        if (prev->next)
            prev->next->prev = prev;

        //D_DEBUG_AT( SurfMan, "  -> freeing %p (prev %p, next %p)\n", chunk, chunk->prev, chunk->next);

        D_MAGIC_CLEAR( chunk );

        SHFREE( manager->shmpool, chunk );
        chunk = prev;
    }

    if (chunk->next  &&  !chunk->next->buffer)
    {
        Chunk *next = chunk->next;

        //D_DEBUG_AT( SurfMan, "  -> merging with next chunk at %d\n", next->offset );

        chunk->length += next->length;

        chunk->next = next->next;
        if (chunk->next)
            chunk->next->prev = chunk;

        D_MAGIC_CLEAR( next );

        SHFREE( manager->shmpool, next );
    }

    return chunk;
}

static Chunk *
occupy_chunk( SurfaceManager *manager, Chunk *chunk, CoreSurfaceAllocation *allocation, int length, int pitch )
{
    D_MAGIC_ASSERT( manager, SurfaceManager );
    D_MAGIC_ASSERT( chunk, Chunk );
    D_MAGIC_ASSERT( allocation, CoreSurfaceAllocation );
    D_MAGIC_ASSERT( allocation->buffer, CoreSurfaceBuffer );

    if (allocation->buffer->policy == CSP_VIDEOONLY)
        manager->m_avail[chunk->poolno] -= length;

    chunk = split_chunk( manager, chunk, length );
    if (!chunk)
        return NULL;

    D_DEBUG_AT( SurfMan, "Allocating %d bytes at offset %d.\n", chunk->length, chunk->offset );

    chunk->allocation = allocation;
    chunk->buffer     = allocation->buffer;
    chunk->pitch  = pitch;

    manager->m_min_toleration[chunk->poolno]++;

    return chunk;
}

