/***************************************************************************/
/*                                                                         */
/*  ftbitmap.c                                                             */
/*                                                                         */
/*    FreeType utility functions for bitmaps (body).                       */
/*                                                                         */
/*  Copyright 2004, 2005, 2006, 2007, 2008, 2009 by                        */
/*  David Turner, Robert Wilhelm, and Werner Lemberg.                      */
/*                                                                         */
/*  This file is part of the FreeType project, and may only be used,       */
/*  modified, and distributed under the terms of the FreeType project      */
/*  license, LICENSE.TXT.  By continuing to use, modify, or distribute     */
/*  this file you indicate that you have read the license and              */
/*  understand and accept it fully.                                        */
/*                                                                         */
/***************************************************************************/


#include <ft2build.h>
#include FT_BITMAP_H
#include FT_IMAGE_H
#include FT_INTERNAL_OBJECTS_H


  static
  const FT_Bitmap  null_bitmap = { 0, 0, 0, 0, 0, 0, 0, 0 };

  typedef enum  FT_Pixel_From_Outline_
  {
    FT_PIXEL_OUT_OUTLINE = 0,
    FT_PIXEL_THRESHOLD  = 128,
    FT_PIXEL_IN_OUTLINE = 129,
    FT_PIXEL_ON_OUTLINE = 255,

    FT_PIXEL_FROM_OUTLINE_MAX      /* do not remove */

  } FT_Pixel_From_Outline;  /* adding for bd_subtitle */

  /* documentation is in ftbitmap.h */

  FT_EXPORT_DEF( void )
  FT_Bitmap_New( FT_Bitmap  *abitmap )
  {
    *abitmap = null_bitmap;
  }


  /* documentation is in ftbitmap.h */

  FT_EXPORT_DEF( FT_Error )
  FT_Bitmap_Copy( FT_Library        library,
                  const FT_Bitmap  *source,
                  FT_Bitmap        *target)
  {
    FT_Memory  memory = library->memory;
    FT_Error   error  = FT_Err_Ok;
    FT_Int     pitch  = source->pitch;
    FT_ULong   size;


    if ( source == target )
      return FT_Err_Ok;

    if ( source->buffer == NULL )
    {
      *target = *source;

      return FT_Err_Ok;
    }

    if ( pitch < 0 )
      pitch = -pitch;
    size = (FT_ULong)( pitch * source->rows );

    if ( target->buffer )
    {
      FT_Int    target_pitch = target->pitch;
      FT_ULong  target_size;


      if ( target_pitch < 0  )
        target_pitch = -target_pitch;
      target_size = (FT_ULong)( target_pitch * target->rows );

      if ( target_size != size )
        (void)FT_QREALLOC( target->buffer, target_size, size );
    }
    else
      (void)FT_QALLOC( target->buffer, size );

    if ( !error )
    {
      unsigned char *p;


      p = target->buffer;
      *target = *source;
      target->buffer = p;

      FT_MEM_COPY( target->buffer, source->buffer, size );
    }

    return error;
  }


  static FT_Error
  ft_bitmap_assure_buffer( FT_Memory   memory,
                           FT_Bitmap*  bitmap,
                           FT_UInt     xpixels,
                           FT_UInt     ypixels )
  {
    FT_Error        error;
    int             pitch;
    int             new_pitch;
    FT_UInt         bpp;
    FT_Int          i, width, height;
    unsigned char*  buffer;


    width  = bitmap->width;
    height = bitmap->rows;
    pitch  = bitmap->pitch;
    if ( pitch < 0 )
      pitch = -pitch;

    switch ( bitmap->pixel_mode )
    {
    case FT_PIXEL_MODE_MONO:
      bpp       = 1;
      new_pitch = ( width + xpixels + 7 ) >> 3;
      break;
    case FT_PIXEL_MODE_GRAY2:
      bpp       = 2;
      new_pitch = ( width + xpixels + 3 ) >> 2;
      break;
    case FT_PIXEL_MODE_GRAY4:
      bpp       = 4;
      new_pitch = ( width + xpixels + 1 ) >> 1;
      break;
    case FT_PIXEL_MODE_GRAY:
    case FT_PIXEL_MODE_LCD:
    case FT_PIXEL_MODE_LCD_V:
      bpp       = 8;
      new_pitch = ( width + xpixels );
      break;
    default:
      return FT_Err_Invalid_Glyph_Format;
    }

    /* if no need to allocate memory */
    if ( ypixels == 0 && new_pitch <= pitch )
    {
      /* zero the padding */
      FT_Int  bit_width = pitch * 8;
      FT_Int  bit_last  = ( width + xpixels ) * bpp;


      if ( bit_last < bit_width )
      {
        FT_Byte*  line  = bitmap->buffer + ( bit_last >> 3 );
        FT_Byte*  end   = bitmap->buffer + pitch;
        FT_Int    shift = bit_last & 7;
        FT_UInt   mask  = 0xFF00U >> shift;
        FT_Int    count = height;


        for ( ; count > 0; count--, line += pitch, end += pitch )
        {
          FT_Byte*  write = line;


          if ( shift > 0 )
          {
            write[0] = (FT_Byte)( write[0] & mask );
            write++;
          }
          if ( write < end )
            FT_MEM_ZERO( write, end-write );
        }
      }

      return FT_Err_Ok;
    }

    if ( FT_QALLOC_MULT( buffer, new_pitch, bitmap->rows + ypixels ) )
      return error;

    if ( bitmap->pitch > 0 )
    {
      FT_Int  len = ( width * bpp + 7 ) >> 3;


      for ( i = 0; i < bitmap->rows; i++ )
        FT_MEM_COPY( buffer + new_pitch * ( ypixels + i ),
                     bitmap->buffer + pitch * i, len );
    }
    else
    {
      FT_Int  len = ( width * bpp + 7 ) >> 3;


      for ( i = 0; i < bitmap->rows; i++ )
        FT_MEM_COPY( buffer + new_pitch * i,
                     bitmap->buffer + pitch * i, len );
    }

    FT_FREE( bitmap->buffer );
    bitmap->buffer = buffer;

    if ( bitmap->pitch < 0 )
      new_pitch = -new_pitch;

    /* set pitch only, width and height are left untouched */
    bitmap->pitch = new_pitch;

    return FT_Err_Ok;
  }


  /* documentation is in ftbitmap.h */

  FT_EXPORT_DEF( FT_Error )
  FT_Bitmap_Embolden( FT_Library  library,
                      FT_Bitmap*  bitmap,
                      FT_Pos      xStrength,
                      FT_Pos      yStrength )
  {
    FT_Error        error;
    unsigned char*  p;
    FT_Int          i, x, y, pitch;
    FT_Int          xstr, ystr;


    if ( !library )
      return FT_Err_Invalid_Library_Handle;

    if ( !bitmap || !bitmap->buffer )
      return FT_Err_Invalid_Argument;

    xstr = FT_PIX_ROUND( xStrength ) >> 6;
    ystr = FT_PIX_ROUND( yStrength ) >> 6;

    if ( xstr == 0 && ystr == 0 )
      return FT_Err_Ok;
    else if ( xstr < 0 || ystr < 0 )
      return FT_Err_Invalid_Argument;

    switch ( bitmap->pixel_mode )
    {
    case FT_PIXEL_MODE_GRAY2:
    case FT_PIXEL_MODE_GRAY4:
      {
        FT_Bitmap  tmp;
        FT_Int     align;


        if ( bitmap->pixel_mode == FT_PIXEL_MODE_GRAY2 )
          align = ( bitmap->width + xstr + 3 ) / 4;
        else
          align = ( bitmap->width + xstr + 1 ) / 2;

        FT_Bitmap_New( &tmp );

        error = FT_Bitmap_Convert( library, bitmap, &tmp, align );
        if ( error )
          return error;

        FT_Bitmap_Done( library, bitmap );
        *bitmap = tmp;
      }
      break;

    case FT_PIXEL_MODE_MONO:
      if ( xstr > 8 )
        xstr = 8;
      break;

    case FT_PIXEL_MODE_LCD:
      xstr *= 3;
      break;

    case FT_PIXEL_MODE_LCD_V:
      ystr *= 3;
      break;
    }

    error = ft_bitmap_assure_buffer( library->memory, bitmap, xstr, ystr );
    if ( error )
      return error;

    pitch = bitmap->pitch;
    if ( pitch > 0 )
      p = bitmap->buffer + pitch * ystr;
    else
    {
      pitch = -pitch;
      p = bitmap->buffer + pitch * ( bitmap->rows - 1 );
    }

    /* for each row */
    for ( y = 0; y < bitmap->rows ; y++ )
    {
      /*
       * Horizontally:
       *
       * From the last pixel on, make each pixel or'ed with the
       * `xstr' pixels before it.
       */
      for ( x = pitch - 1; x >= 0; x-- )
      {
        unsigned char tmp;


        tmp = p[x];
        for ( i = 1; i <= xstr; i++ )
        {
          if ( bitmap->pixel_mode == FT_PIXEL_MODE_MONO )
          {
            p[x] |= tmp >> i;

            /* the maximum value of 8 for `xstr' comes from here */
            if ( x > 0 )
              p[x] |= p[x - 1] << ( 8 - i );

#if 0
            if ( p[x] == 0xff )
              break;
#endif
          }
          else
          {
            if ( x - i >= 0 )
            {
              if ( p[x] + p[x - i] > bitmap->num_grays - 1 )
              {
                p[x] = (unsigned char)(bitmap->num_grays - 1);
                break;
              }
              else
              {
                p[x] = (unsigned char)(p[x] + p[x-i]);
                if ( p[x] == bitmap->num_grays - 1 )
                  break;
              }
            }
            else
              break;
          }
        }
      }

      /*
       * Vertically:
       *
       * Make the above `ystr' rows or'ed with it.
       */
      for ( x = 1; x <= ystr; x++ )
      {
        unsigned char*  q;


        q = p - bitmap->pitch * x;
        for ( i = 0; i < pitch; i++ )
          q[i] |= p[i];
      }

      p += bitmap->pitch;
    }

    bitmap->width += xstr;
    bitmap->rows += ystr;

    return FT_Err_Ok;
  }


  /* documentation is in ftbitmap.h */

  FT_EXPORT_DEF( FT_Error )
  FT_Bitmap_Convert( FT_Library        library,
                     const FT_Bitmap  *source,
                     FT_Bitmap        *target,
                     FT_Int            alignment )
  {
    FT_Error   error = FT_Err_Ok;
    FT_Memory  memory;


    if ( !library )
      return FT_Err_Invalid_Library_Handle;

    memory = library->memory;

    switch ( source->pixel_mode )
    {
    case FT_PIXEL_MODE_MONO:
    case FT_PIXEL_MODE_GRAY:
    case FT_PIXEL_MODE_GRAY2:
    case FT_PIXEL_MODE_GRAY4:
    case FT_PIXEL_MODE_LCD:
    case FT_PIXEL_MODE_LCD_V:
      {
        FT_Int   pad;
        FT_Long  old_size;


        old_size = target->rows * target->pitch;
        if ( old_size < 0 )
          old_size = -old_size;

        target->pixel_mode = FT_PIXEL_MODE_GRAY;
        target->rows       = source->rows;
        target->width      = source->width;

        pad = 0;
        if ( alignment > 0 )
        {
          pad = source->width % alignment;
          if ( pad != 0 )
            pad = alignment - pad;
        }

        target->pitch = source->width + pad;

        if ( target->rows * target->pitch > old_size             &&
             FT_QREALLOC( target->buffer,
                          old_size, target->rows * target->pitch ) )
          return error;
      }
      break;

    default:
      error = FT_Err_Invalid_Argument;
    }

    switch ( source->pixel_mode )
    {
    case FT_PIXEL_MODE_MONO:
      {
        FT_Byte*  s = source->buffer;
        FT_Byte*  t = target->buffer;
        FT_Int    i;


        target->num_grays = 2;

        for ( i = source->rows; i > 0; i-- )
        {
          FT_Byte*  ss = s;
          FT_Byte*  tt = t;
          FT_Int    j;


          /* get the full bytes */
          for ( j = source->width >> 3; j > 0; j-- )
          {
            FT_Int  val = ss[0]; /* avoid a byte->int cast on each line */


            tt[0] = (FT_Byte)( ( val & 0x80 ) >> 7 );
            tt[1] = (FT_Byte)( ( val & 0x40 ) >> 6 );
            tt[2] = (FT_Byte)( ( val & 0x20 ) >> 5 );
            tt[3] = (FT_Byte)( ( val & 0x10 ) >> 4 );
            tt[4] = (FT_Byte)( ( val & 0x08 ) >> 3 );
            tt[5] = (FT_Byte)( ( val & 0x04 ) >> 2 );
            tt[6] = (FT_Byte)( ( val & 0x02 ) >> 1 );
            tt[7] = (FT_Byte)(   val & 0x01 );

            tt += 8;
            ss += 1;
          }

          /* get remaining pixels (if any) */
          j = source->width & 7;
          if ( j > 0 )
          {
            FT_Int  val = *ss;


            for ( ; j > 0; j-- )
            {
              tt[0] = (FT_Byte)( ( val & 0x80 ) >> 7);
              val <<= 1;
              tt   += 1;
            }
          }

          s += source->pitch;
          t += target->pitch;
        }
      }
      break;


    case FT_PIXEL_MODE_GRAY:
    case FT_PIXEL_MODE_LCD:
    case FT_PIXEL_MODE_LCD_V:
      {
        FT_Int    width   = source->width;
        FT_Byte*  s       = source->buffer;
        FT_Byte*  t       = target->buffer;
        FT_Int    s_pitch = source->pitch;
        FT_Int    t_pitch = target->pitch;
        FT_Int    i;


        target->num_grays = 256;

        for ( i = source->rows; i > 0; i-- )
        {
          FT_ARRAY_COPY( t, s, width );

          s += s_pitch;
          t += t_pitch;
        }
      }
      break;


    case FT_PIXEL_MODE_GRAY2:
      {
        FT_Byte*  s = source->buffer;
        FT_Byte*  t = target->buffer;
        FT_Int    i;


        target->num_grays = 4;

        for ( i = source->rows; i > 0; i-- )
        {
          FT_Byte*  ss = s;
          FT_Byte*  tt = t;
          FT_Int    j;


          /* get the full bytes */
          for ( j = source->width >> 2; j > 0; j-- )
          {
            FT_Int  val = ss[0];


            tt[0] = (FT_Byte)( ( val & 0xC0 ) >> 6 );
            tt[1] = (FT_Byte)( ( val & 0x30 ) >> 4 );
            tt[2] = (FT_Byte)( ( val & 0x0C ) >> 2 );
            tt[3] = (FT_Byte)( ( val & 0x03 ) );

            ss += 1;
            tt += 4;
          }

          j = source->width & 3;
          if ( j > 0 )
          {
            FT_Int  val = ss[0];


            for ( ; j > 0; j-- )
            {
              tt[0]  = (FT_Byte)( ( val & 0xC0 ) >> 6 );
              val  <<= 2;
              tt    += 1;
            }
          }

          s += source->pitch;
          t += target->pitch;
        }
      }
      break;


    case FT_PIXEL_MODE_GRAY4:
      {
        FT_Byte*  s = source->buffer;
        FT_Byte*  t = target->buffer;
        FT_Int    i;


        target->num_grays = 16;

        for ( i = source->rows; i > 0; i-- )
        {
          FT_Byte*  ss = s;
          FT_Byte*  tt = t;
          FT_Int    j;


          /* get the full bytes */
          for ( j = source->width >> 1; j > 0; j-- )
          {
            FT_Int  val = ss[0];


            tt[0] = (FT_Byte)( ( val & 0xF0 ) >> 4 );
            tt[1] = (FT_Byte)( ( val & 0x0F ) );

            ss += 1;
            tt += 2;
          }

          if ( source->width & 1 )
            tt[0] = (FT_Byte)( ( ss[0] & 0xF0 ) >> 4 );

          s += source->pitch;
          t += target->pitch;
        }
      }
      break;


    default:
      ;
    }

    return error;
  }


  /* documentation is in ftbitmap.h */

  FT_EXPORT_DEF( FT_Error )
  FT_GlyphSlot_Own_Bitmap( FT_GlyphSlot  slot )
  {
    if ( slot && slot->format == FT_GLYPH_FORMAT_BITMAP   &&
         !( slot->internal->flags & FT_GLYPH_OWN_BITMAP ) )
    {
      FT_Bitmap  bitmap;
      FT_Error   error;


      FT_Bitmap_New( &bitmap );
      error = FT_Bitmap_Copy( slot->library, &slot->bitmap, &bitmap );
      if ( error )
        return error;

      slot->bitmap = bitmap;
      slot->internal->flags |= FT_GLYPH_OWN_BITMAP;
    }

    return FT_Err_Ok;
  }


  /* documentation is in ftbitmap.h */

  FT_EXPORT_DEF( FT_Error )
  FT_Bitmap_Done( FT_Library  library,
                  FT_Bitmap  *bitmap )
  {
    FT_Memory  memory;


    if ( !library )
      return FT_Err_Invalid_Library_Handle;

    if ( !bitmap )
      return FT_Err_Invalid_Argument;

    memory = library->memory;

    FT_FREE( bitmap->buffer );
    *bitmap = null_bitmap;

    return FT_Err_Ok;
  }
#ifndef FE_DISABLE_EDGE_EFFECT_SUPPORT
	/* Extend bitmap buffer and move the glyph to assigned origin;
	   allocate edge buffer */
	static FT_Error
	ft_bitmap_assure_edge_buffer( FT_Memory   memory,
							 FT_Bitmap*  bitmap,
							 FT_UInt	 xpixels,
							 FT_UInt	 ypixels,
							 FT_UInt	 xmove,
							 FT_UInt	 ymove )
	{
	  FT_Error		  error;
	  int			  pitch;
	  int			  new_pitch;
	  FT_UInt		  ppb;
	  FT_Int		  i;
	  FT_Int		  j;
	  unsigned char*  buffer;
  
  
	  pitch = bitmap->pitch;
	  if ( pitch < 0 )
		pitch = -pitch;
  
	  switch ( bitmap->pixel_mode )
	  {
	  case FT_PIXEL_MODE_MONO:
		ppb = 8;
		break;
	  case FT_PIXEL_MODE_GRAY2:
		ppb = 4;
		break;
	  case FT_PIXEL_MODE_GRAY4:
		ppb = 2;
		break;
	  case FT_PIXEL_MODE_GRAY:
	  case FT_PIXEL_MODE_LCD:
	  case FT_PIXEL_MODE_LCD_V:
		ppb = 1;
		break;
	  default:
		return FT_Err_Invalid_Glyph_Format;
	  }
  
	  new_pitch = ( bitmap->width + xpixels + ppb - 1 ) / ppb;
  
	  if ( FT_ALLOC( buffer, (new_pitch * ( bitmap->rows + ypixels ) ) * 2 ) )
		return error;
  
	  if ( bitmap->pitch > 0 )
	  {
		for ( i = 0; i < bitmap->rows; i++ )
		  FT_MEM_COPY( buffer + new_pitch * ( ymove + i ) + ( xmove / ppb ),
					   bitmap->buffer + pitch * i, pitch );
	  }
	  else
	  {
		for ( i = 0; i < bitmap->rows; i++ )
		  FT_MEM_COPY( buffer + new_pitch * ( ymove - ypixels + i ) + ( ( xmove - xpixels ) / ppb ),
					   bitmap->buffer + pitch * i, pitch );
	  }
  
	  if ( ( xmove / ppb ) * ppb != xmove )
	  { /* Need shift */
		FT_Int			shift;
  
		shift = ( xmove - ( ( xmove / ppb ) * ppb ) ) * (8 / ppb);
  
		for ( i = 0; i < bitmap->rows + ypixels; i++ )
		{
		  for ( j = new_pitch - 1; j > 0; j-- )
			buffer[new_pitch * i + j] = ( buffer[new_pitch * i + j - 1] << (8 - shift) ) |
										( buffer[new_pitch * i + j] >> shift);
		  buffer[new_pitch * i] >>= shift;
		}
	  }
  
	  FT_FREE( bitmap->buffer );
	  bitmap->buffer = buffer;
  
	  if ( bitmap->pitch < 0 )
		new_pitch = -new_pitch;
  
	  bitmap->pitch = new_pitch;
	  bitmap->rows += ypixels;
  
	  return FT_Err_Ok;
	}
  
  
  
	FT_EXPORT_DEF( FT_Error )
	FT_Bitmap_Shadow( FT_Library  library,
						FT_Bitmap*	bitmap,
						FT_Pos		dx,
						FT_Pos		dy )
	{
	  FT_Error		  error;
	  unsigned char*  p;
	  unsigned char*  pedge;
	  FT_Int		  x, y, pitch;
  
	  if ( !library )
		return FT_Err_Invalid_Library_Handle;
  
	  if ( !bitmap || !bitmap->buffer )
		return FT_Err_Invalid_Argument;
  
	  if ( (dx == 0) && (dy == 0) )
		return FT_Err_Ok;
  
	  switch ( bitmap->pixel_mode )
	  {
	  case FT_PIXEL_MODE_GRAY2:
	  case FT_PIXEL_MODE_GRAY4:
		{
		  FT_Bitmap  tmp;
		  FT_Int	 align;
  
  
		  if ( bitmap->pixel_mode == FT_PIXEL_MODE_GRAY2 )
			align = ( bitmap->width + dx + 3 ) / 4;
		  else
			align = ( bitmap->width + dx + 1 ) / 2;
  
		  FT_Bitmap_New( &tmp );
		  error = FT_Bitmap_Convert( library, bitmap, &tmp, align );
  
		  if ( error )
			return error;
  
		  FT_Bitmap_Done( library, bitmap );
		  *bitmap = tmp;
		}
		break;
  
	  case FT_PIXEL_MODE_MONO:
		if ( dx > 8 )
		  dx = 8;
		break;
  
	  case FT_PIXEL_MODE_LCD:
		dx *= 3;
		break;
  
	  case FT_PIXEL_MODE_LCD_V:
		dy *= 3;
		break;
	  }
  
	  error = ft_bitmap_assure_edge_buffer( library->memory, bitmap, 
				  (dx > 0)? dx: -dx, 
				  (dy > 0)? dy: -dy, 
				  (dx > 0)? 0: -dx, 
				  (dy > 0)? 0: -dy );
	  if ( error )
		return error;
  
	  pitch = bitmap->pitch;
	  p = bitmap->buffer;
  
	  pedge = p + (pitch * bitmap->rows);
	  FT_MEM_SET(pedge, 0, pitch * bitmap->rows);
  
	  if ( bitmap->pixel_mode == FT_PIXEL_MODE_MONO )
	  {
		for ( y = 0; y < bitmap->rows ; y++ )
		{
		  for ( x = pitch - 1; x >= 0; x-- )
		  {
			if ( (y + dy >= 0) && (y + dy < bitmap->rows) )
			{
			  if ( dx > 0 )
			  {
				pedge[dy * pitch + x] = (p[x] >> dx);
				if ( x > 0 )
				  pedge[x] |= p[x - 1] << ( 8 - dx );
			  }
			  else
			  {
				pedge[dy * pitch + x] = (p[x] << -dx);
				if ( x < pitch - 1 )
				  pedge[x] |= p[x + 1] >> ( 8 + dx );
			  }
  
			  pedge[dy * pitch + x] ^= p[dy * pitch + x];
			}
		  }
  
		  p += pitch;
		  pedge += pitch;
		}
	  }
	  else
	  {
		for ( y = 0; y < bitmap->rows ; y++ )
		{
		  for ( x = 0; x < pitch ; x++ )
		  {
			if (p[x] && (bitmap->num_grays != p[dy * pitch + x + dx]))
			{
			  if ((y + dy >= 0) && (y + dy < bitmap->rows) && 
				  (x + dx >= 0) && (x + dx < pitch))
			  {
				pedge[dy * pitch + x + dx] = 
					p[x] * (bitmap->num_grays - p[dy * pitch + x + dx]) / bitmap->num_grays;
			  }
			}
		  }
  
		  p += pitch;
		  pedge += pitch;
		}
	  }
  
	  bitmap->width += (dx > 0)? dx: -dx;
  
	  return FT_Err_Ok;
	}
  


	FT_EXPORT_DEF( FT_Error )
	FT_Bitmap_Raised( FT_Library  library,
						FT_Bitmap*	bitmap,
						FT_Pos		dx,
						FT_Pos		dy )
	{
	  FT_Error		  error;
	  unsigned char*  p;
	  unsigned char*  pedge;
	  FT_Int		  dxa, dya;
	  FT_Int		  i, x, y, pitch;
  
  
	  if ( !library )
		return FT_Err_Invalid_Library_Handle;
  
	  if ( !bitmap || !bitmap->buffer )
		return FT_Err_Invalid_Argument;
  
	  if ( (dx == 0) && (dy == 0) )
		return FT_Err_Ok;
  
	  if ( dx < 0 )
		dxa = -dx;
	  else
		dxa = dx;
	  if ( dy < 0 )
		dya = -dy;
	  else
		dya = dy;
  
	  switch ( bitmap->pixel_mode )
	  {
	  case FT_PIXEL_MODE_GRAY2:
	  case FT_PIXEL_MODE_GRAY4:
		{
		  FT_Bitmap  tmp;
		  FT_Int	 align;
  
  
		  if ( bitmap->pixel_mode == FT_PIXEL_MODE_GRAY2 )
			align = ( bitmap->width + dxa + 3 ) / 4;
		  else
			align = ( bitmap->width + dxa + 1 ) / 2;
  
		  FT_Bitmap_New( &tmp );
		  error = FT_Bitmap_Convert( library, bitmap, &tmp, align );
  
		  if ( error )
			return error;
  
		  FT_Bitmap_Done( library, bitmap );
		  *bitmap = tmp;
		}
		break;
  
	  case FT_PIXEL_MODE_MONO:
		if ( dx > 8 )
		  dx = 8;
		break;
  
	  case FT_PIXEL_MODE_LCD:
		dx *= 3;
		break;
  
	  case FT_PIXEL_MODE_LCD_V:
		dy *= 3;
		break;
	  }
  
	  error = ft_bitmap_assure_edge_buffer( library->memory, bitmap, 
				  (dx > 0)? dx: dxa, 
				  (dy > 0)? dy: dya, 
				  (dx > 0)? 0: dxa, 
				  (dy > 0)? 0: dya );
	  if ( error )
		return error;
  
	  pitch = bitmap->pitch;
	  p = bitmap->buffer;
  
	  pedge = p + (pitch * bitmap->rows);
	  FT_MEM_SET(pedge, 0, pitch * bitmap->rows);
  
	  if ( bitmap->pixel_mode == FT_PIXEL_MODE_MONO )
	  {
		for ( y = 0; y < (bitmap->rows - dya) ; y++ )
		{
		  for ( x = pitch - 1; x >= 0; x-- )
		  {
			if ( dx > 0 )
			{
			  for ( i = dx; i > 0; i-- )
			  {
				FT_Int iy = dy * i / dx;
				FT_Int idx = iy * pitch + x;
	
				pedge[idx] |= (p[x] >> i);
				if ( x < pitch - 1 )
				  pedge[idx + 1] |= p[x] << ( 8 - i );
			  }
			}
			else
			{
			  for ( i = 0; i < dxa; i++ )
			  {
				FT_Int iy = dya * i / dxa;
				FT_Int idx = iy * pitch + x;
	
				pedge[idx] |= (p[dya * pitch + x] << i);
				if ( x > 0 )
				  pedge[idx - 1] |= p[dya * pitch + x] >> ( 8 - i );
			  }
			}
		  }
  
		  p += pitch;
		  pedge += pitch;
		}
  
		p = bitmap->buffer;
		pedge = p + (pitch * bitmap->rows);
		for ( y = 0; y < bitmap->rows ; y++ )
		{
		  for ( x = pitch - 1; x >= 0; x-- )
		  {
			pedge[x] ^= p[x];
		  }
  
		  p += pitch;
		  pedge += pitch;
		}
	  }
	  else
	  {
		for ( y = 0; y < (bitmap->rows - dya) ; y++ )
		{
		  for ( x = 0; x < (pitch - dxa) ; x++ )
		  {
			if ((dx > 0) && p[x])
			{
			  unsigned char gray = p[x];
			  unsigned char gray2 = p[x] * 3 / 4;
	
			  for ( i = dx; i > 0; i-- )
			  {
				FT_Int iy = dy * i / dx;
				FT_Int idx = iy * pitch + x + i;
	
				if ((iy * dx) == (dy * i))
				{
				  pedge[idx] |= gray;
				}
				else
				{
				  pedge[idx] |= gray2;
				  pedge[idx + pitch] |= gray2;
				}
			  }
			}
			else if ((dx < 0) && p[dya * pitch + x + dxa])
			{
			  unsigned char gray = p[dya * pitch + x + dxa];
			  unsigned char gray2 = gray * 3 / 4;
	
			  for ( i = 0; i < dxa; i++ )
			  {
				FT_Int iy = dya * i / dxa;
				FT_Int idx = iy * pitch + x + i;
   
				if ((iy * dxa) == (dya * i))
				{
				  pedge[idx] |= gray;
				}
				else
				{
				  pedge[idx] |= gray2;
				  pedge[idx + pitch] |= gray2;
				}
			  }
			}
		  }
	
		  p += pitch;
		  pedge += pitch;
		}
  
		p = bitmap->buffer;
		pedge = p + (pitch * bitmap->rows);
		for ( y = 0; y < bitmap->rows; y++ )
		{
		  for ( x = 0; x < pitch; x++ )
		  {
			if (pedge[x] && p[x])
			{
				pedge[x] = pedge[x] * (bitmap->num_grays - p[x]) / (bitmap->num_grays);
			}
		  }
		  p += pitch;
		  pedge += pitch;
		}
	  }
  
	  bitmap->width += (dx > 0)? dx: dxa;
  
	  return FT_Err_Ok;
	}
  
  
  
	FT_EXPORT_DEF( FT_Error )
	FT_Bitmap_Uniform( FT_Library  library,
						FT_Bitmap*	bitmap,
						FT_Pos		width )
	{
	  FT_Error		  error;
	  unsigned char*  p;
	  unsigned char*  q;
	  unsigned char*  pedge;
	  FT_Int		  i, j, x, y, pitch;
  
  
	  if ( !library )
		return FT_Err_Invalid_Library_Handle;
  
	  if ( !bitmap || !bitmap->buffer )
		return FT_Err_Invalid_Argument;
  
	  if ( width <= 0 )
		return FT_Err_Ok;
  
  
	  switch ( bitmap->pixel_mode )
	  {
	  case FT_PIXEL_MODE_GRAY2:
	  case FT_PIXEL_MODE_GRAY4:
		{
		  FT_Bitmap  tmp;
		  FT_Int	 align;
  
  
		  if ( bitmap->pixel_mode == FT_PIXEL_MODE_GRAY2 )
			align = ( bitmap->width + width + 3 ) / 4;
		  else
			align = ( bitmap->width + width + 1 ) / 2;
  
		  FT_Bitmap_New( &tmp );
		  error = FT_Bitmap_Convert( library, bitmap, &tmp, align );
  
		  if ( error )
			return error;
  
		  FT_Bitmap_Done( library, bitmap );
		  *bitmap = tmp;
		}
		break;
  
	  case FT_PIXEL_MODE_MONO:
		if ( width > 8 )
		  width = 8;
		break;
  
	  case FT_PIXEL_MODE_LCD:
		width *= 3;
		break;
  
	  case FT_PIXEL_MODE_LCD_V:
		width *= 3;
		break;
	  }
  
	  error = ft_bitmap_assure_edge_buffer( library->memory, bitmap, width * 2, width * 2, width, width );
	  if ( error )
		return error;
  
	  pitch = bitmap->pitch;
	  p = bitmap->buffer;
  
	  pedge = p + (pitch * bitmap->rows);
	  FT_MEM_SET(pedge, 0, pitch * bitmap->rows);
  
	  if ( bitmap->pixel_mode == FT_PIXEL_MODE_MONO )
	  {
		for ( y = 0; y < bitmap->rows ; y++ )
		{
		  /* Horizontally */
		  for ( x = pitch - 1; x >= 0; x-- )
		  {
			for ( i = 1; i <= width; i++ )
			{
			  pedge[x] |= (p[x] >> i) | (p[x] << i);
			  if ( x > 0 )
				pedge[x] |= p[x - 1] << ( 8 - i );
			  if ( x < pitch - 1 )
				pedge[x] |= p[x + 1] >> ( 8 - i );
			}
		  }
		  /* Vertically */
		  for ( x = 1; x <= width; x++ )
		  {
			if ( y - x >= 0 )
			{
			  q = p - pitch * x;
			  for ( i = 0; i < pitch; i++ )
				pedge[i] |= q[i];
			}
			if ( y + x < bitmap->rows )
			{
			  q = p + pitch * x;
			  for ( i = 0; i < pitch; i++ )
				pedge[i] |= q[i];
			}
		  }
		  /* Take out the font body */
		  for ( x = pitch - 1; x >= 0; x-- )
		  {
			pedge[x] ^= p[x];
		  }
  
		  p += pitch;
		  pedge += pitch;
		}
	  }
	  else
	  {
		p += width * pitch + width;
  
		/* fill the value of each pixel to its surrounding edge pixels if it's larger;
		   this ensures that each edge pixel gets the largest value */
		for ( y = 0; y < (bitmap->rows - (2 * width)) ; y++ )
		{
		  for ( x = 0; x < (pitch - (2 * width)) ; x++ )
		  {
			unsigned char gray = p[x];
  
			for ( i = 0; i <= (2 * width); i++ )
			{
			  unsigned char *ptemp = &pedge[i * pitch + x];
  
			  for ( j = 0; j <= (2 * width); j++ )
			  {
				ptemp[j] |= gray;
			  }
			}
		  }
  
		  p += pitch;
		  pedge += pitch;
		}
  
		p = bitmap->buffer;
		pedge = p + (pitch * bitmap->rows);
		for ( y = 0; y < bitmap->rows; y++ )
		{
		  for ( x = 0; x < pitch; x++ )
		  {
			if (pedge[x] && p[x])
			{
			  pedge[x] = pedge[x] * (bitmap->num_grays - p[x]) / (bitmap->num_grays);
			}
		  }
		  p += pitch;
		  pedge += pitch;
		}
	  }
  
	  bitmap->width += 2 * width;
  
	  return FT_Err_Ok;
	}
   
  
	FT_EXPORT_DEF( FT_Error )
  FT_Bitmap_Uniform_CCI_ORN( FT_Library  library,
                      FT_Bitmap*  bitmap,
                      FT_Pos      width )
  {
    FT_Error        error;
    unsigned char*  p;
    unsigned char*  q;
    unsigned char*  pedge;
    FT_Int          i, j, x, y, pitch;
    /*for ORN */
    FT_Int          buf_count_old;
    unsigned char*  buffer_edge_old;
    FT_Memory 	  memory = library->memory;


    if ( !library )
      return FT_Err_Invalid_Library_Handle;

    if ( !bitmap || !bitmap->buffer )
      return FT_Err_Invalid_Argument;

    if ( width <= 0 )
      return FT_Err_Ok;


    switch ( bitmap->pixel_mode )
    {
    case FT_PIXEL_MODE_GRAY2:
    case FT_PIXEL_MODE_GRAY4:
      {
        FT_Bitmap  tmp;
        FT_Int     align;


        if ( bitmap->pixel_mode == FT_PIXEL_MODE_GRAY2 )
          align = ( bitmap->width + width + 3 ) / 4;
        else
          align = ( bitmap->width + width + 1 ) / 2;

        FT_Bitmap_New( &tmp );
        error = FT_Bitmap_Convert( library, bitmap, &tmp, align );

        if ( error )
          return error;

        FT_Bitmap_Done( library, bitmap );
        *bitmap = tmp;
      }
      break;

    case FT_PIXEL_MODE_MONO:
      if ( width > 8 )
        width = 8;
      break;

    case FT_PIXEL_MODE_LCD:
      width *= 3;
      break;

    case FT_PIXEL_MODE_LCD_V:
      width *= 3;
      break;
    }

    buf_count_old = 1;      /*bitmap->buf_count*/
    error = ft_bitmap_assure_edge_buffer( library->memory, bitmap, width * 2, width * 2, width, width );
    if ( error )
      return error;

    pitch = bitmap->pitch;
    p = bitmap->buffer;

    pedge = p + (pitch * bitmap->rows);

    if (buf_count_old == 2)
    {
	   if ( FT_ALLOC( buffer_edge_old,	bitmap->pitch * (bitmap->rows)))
			 return error;
	   
	   if ( bitmap->pitch > 0 )
	   {
		 for ( i = 0; i < bitmap->rows; i++ )
		   FT_MEM_COPY( buffer_edge_old + bitmap->pitch * i,
						pedge + pitch * i, pitch);
	   }
	   else
	   {
		 for ( i = 0; i < bitmap->rows; i++ )
		   FT_MEM_COPY( buffer_edge_old +  bitmap->pitch * i,
						bitmap->buffer + pitch * i, pitch );
	   }
    }
    
    FT_MEM_SET(pedge, 0, pitch * bitmap->rows);

    if ( bitmap->pixel_mode == FT_PIXEL_MODE_MONO )
    {
      for ( y = 0; y < bitmap->rows ; y++ )
      {
        /* Horizontally */
        for ( x = pitch - 1; x >= 0; x-- )
        {
          for ( i = 1; i <= width; i++ )
          {
            pedge[x] |= (p[x] >> i) | (p[x] << i);
            if ( x > 0 )
              pedge[x] |= p[x - 1] << ( 8 - i );
            if ( x < pitch - 1 )
              pedge[x] |= p[x + 1] >> ( 8 - i );
          }
        }
        /* Vertically */
        for ( x = 1; x <= width; x++ )
        {
          if ( y - x >= 0 )
          {
            q = p - pitch * x;
            for ( i = 0; i < pitch; i++ )
              pedge[i] |= q[i];
          }
          if ( y + x < bitmap->rows )
          {
            q = p + pitch * x;
            for ( i = 0; i < pitch; i++ )
              pedge[i] |= q[i];
          }
        }
        /* Take out the font body */
        for ( x = pitch - 1; x >= 0; x-- )
        {
          pedge[x] ^= p[x];
        }

        p += pitch;
        pedge += pitch;
      }
    }
    else
    {
      p += width * pitch + width;

      /* fill the value of each pixel to its surrounding edge pixels if it's larger;
         this ensures that each edge pixel gets the largest value */
      for ( y = 0; y < (bitmap->rows - (2 * width)) ; y++ )
      {
        for ( x = 0; x < (pitch - (2 * width)) ; x++ )
        {
          unsigned char gray = p[x];

          for ( i = 0; i <= (2 * width); i++ )
          {
            unsigned char *ptemp = &pedge[i * pitch + x];

            for ( j = 0; j <= (2 * width); j++ )
            {
              ptemp[j] |= gray;
            }
          }
        }

        p += pitch;
        pedge += pitch;
      }

      p = bitmap->buffer;
      pedge = p + (pitch * bitmap->rows);
      for ( y = 0; y < bitmap->rows; y++ )
      {
        for ( x = 0; x < pitch; x++ )
        {
          if (pedge[x] && p[x])
          {
            pedge[x] = pedge[x] * (bitmap->num_grays - p[x]) / (bitmap->num_grays);
          }
        }
        p += pitch;
        pedge += pitch;
      }
    }

    if (buf_count_old == 2)
    {
        p = bitmap->buffer; 
        pedge = p + (pitch * bitmap->rows);
        for ( y = 0; y < bitmap->rows ; y++ )
        {
            for ( x = 0; x < pitch ; x++ )
            {
                pedge[x] |= buffer_edge_old[x];
            }
            buffer_edge_old += pitch;
            pedge += pitch;
        }
    }

    bitmap->width += 2 * width;

    return FT_Err_Ok;
  }



  FT_EXPORT_DEF( FT_Error )
	FT_Bitmap_Outline( FT_Library  library,
						FT_Bitmap*	bitmap,
						FT_Pos		width )
	{
	  FT_Memory 	  memory = library->memory;
	  FT_Error		  error;
	  int			  new_pitch;
	  int			  new_rows;
	  FT_UInt		  ppb;
	  unsigned char*  buffer;
	  unsigned char*  p;
	  unsigned char*  p2;
	  unsigned char*  pedge;
	  FT_Int		  i, j, x, y, pitch;
  
  
	  if ( !library )
		return FT_Err_Invalid_Library_Handle;
  
	  if ( !bitmap || !bitmap->buffer )
		return FT_Err_Invalid_Argument;
  
	  if ( width <= 0 )
		return FT_Err_Ok;
  
  
	  switch ( bitmap->pixel_mode )
	  {
	  case FT_PIXEL_MODE_GRAY2:
	  case FT_PIXEL_MODE_GRAY4:
		{
		  FT_Bitmap  tmp;
		  FT_Int	 align;
  
  
		  if ( bitmap->pixel_mode == FT_PIXEL_MODE_GRAY2 )
			align = ( bitmap->width + width + 3 ) / 4;
		  else
			align = ( bitmap->width + width + 1 ) / 2;
  
		  FT_Bitmap_New( &tmp );
		  error = FT_Bitmap_Convert( library, bitmap, &tmp, align );
  
		  if ( error )
			return error;
  
		  FT_Bitmap_Done( library, bitmap );
		  *bitmap = tmp;
		}
		break;
  
	  case FT_PIXEL_MODE_MONO:
		if ( width > 8 )
		  width = 8;
		break;
  
	  case FT_PIXEL_MODE_LCD:
		width *= 3;
		break;
  
	  case FT_PIXEL_MODE_LCD_V:
		width *= 3;
		break;
	  }
  
	  pitch = bitmap->pitch;
	  if ( pitch < 0 )
		pitch = -pitch;
  
	  switch ( bitmap->pixel_mode )
	  {
	  case FT_PIXEL_MODE_MONO:
		ppb = 8;
		break;
	  case FT_PIXEL_MODE_GRAY2:
		ppb = 4;
		break;
	  case FT_PIXEL_MODE_GRAY4:
		ppb = 2;
		break;
	  case FT_PIXEL_MODE_GRAY:
	  case FT_PIXEL_MODE_LCD:
	  case FT_PIXEL_MODE_LCD_V:
		ppb = 1;
		break;
	  default:
		return FT_Err_Invalid_Glyph_Format;
	  }
  
	  new_pitch = ( bitmap->width + 2 * width + ppb - 1 ) / ppb;
	  new_rows = bitmap->rows + 2 * width;
  
	  if ( FT_ALLOC( buffer, new_pitch * new_rows) )
		return error;
	  FT_MEM_SET(buffer, 0, new_pitch * new_rows);
  
	  pitch = bitmap->pitch;
	  p = bitmap->buffer;
  
	  pedge = buffer;
  
	  if ( bitmap->pixel_mode == FT_PIXEL_MODE_MONO )
	  {
		p += (bitmap->rows - 1) * pitch;
		pedge += (bitmap->rows - 1) * new_pitch;
  
		for ( y = bitmap->rows - 1; y >= 0 ; y-- )
		{
		  /* Horizontally */
		  for ( x = pitch - 1; x >= 0; x-- )
		  {
			for ( i = 0; i <= 2 * width; i++ )
			{
			  pedge[x] |= (p[x] >> i);
			  if ( x + 1 < new_pitch )
				pedge[x + 1] |= (p[x] << ( 8 - i ));
			}
		  }
		  /* Vertically */
		  for ( x = 1; x <= 2 * width; x++ )
		  {
			for ( i = 0; i < new_pitch; i++ )
			{
			  pedge[x * new_pitch + i] |= pedge[i];
			}
		  }
  
		  p -= pitch;
		  pedge -= new_pitch;
		}
  
		p = bitmap->buffer;
		pedge = buffer + width * new_pitch;
		for ( y = 0; y < bitmap->rows ; y++ )
		{
		  for ( x = pitch - 1; x >= 0; x-- )
		  {
			if ( x + 1 < new_pitch )
			  pedge[x + 1] &= ~(p[x] << ( 8 - width ));
		  }
  
		  p += pitch;
		  pedge += new_pitch;
		}
	  }
	  else
	  {
		p2 = p;
  
		/* fill the value of each pixel to its surrounding edge pixels if it's larger;
		   this ensures that each edge pixel gets the largest value */
		for ( y = 0; y < bitmap->rows ; y++ )
		{
		  for ( x = 0; x < pitch ; x++ )
		  {
			unsigned char gray = p[x];
	
			for ( i = 0; i <= (2 * width); i++ )
			{
			  unsigned char *ptemp = &pedge[i * new_pitch + x];
	
			  for ( j = 0; j <= (2 * width); j++ )
			  {
				ptemp[j] |= gray;
			  }
			}
		  }
  
		  /* cancel out the font body from edge effect - after the edge pixels get their
			 largest value */
		  if (y >= width)
		  {
			for ( x = width; x < (pitch + width); x++ )
			{
			  if (pedge[x] && p2[x - width])
			  {
				pedge[x] = pedge[x] * (bitmap->num_grays - p2[x - width]) / (bitmap->num_grays);
			  }
			}
			p2 += pitch;
		  }
	
		  p += pitch;
		  pedge += new_pitch;
		}
	  }
  
	  FT_FREE( bitmap->buffer );
	  bitmap->buffer = buffer;
  
	  bitmap->pitch = new_pitch;
	  bitmap->rows = new_rows;
	  bitmap->width += 2 * width;
  
	  return FT_Err_Ok;
	}
  
#endif

/* END */
