/***************************************************************************/
/*                                                                         */
/*  ftsynth.c                                                              */
/*                                                                         */
/*    FreeType synthesizing code for emboldening and slanting (body).      */
/*                                                                         */
/*  Copyright 2000-2001, 2002, 2003, 2004, 2005, 2006 by                   */
/*  David Turner, Robert Wilhelm, and Werner Lemberg.                      */
/*                                                                         */
/*  This file is part of the FreeType project, and may only be used,       */
/*  modified, and distributed under the terms of the FreeType project      */
/*  license, LICENSE.TXT.  By continuing to use, modify, or distribute     */
/*  this file you indicate that you have read the license and              */
/*  understand and accept it fully.                                        */
/*                                                                         */
/***************************************************************************/


#include <ft2build.h>
#include FT_SYNTHESIS_H
#include FT_INTERNAL_OBJECTS_H
#include FT_OUTLINE_H
#include FT_BITMAP_H


  /*************************************************************************/
  /*************************************************************************/
  /****                                                                 ****/
  /****   EXPERIMENTAL OBLIQUING SUPPORT                                ****/
  /****                                                                 ****/
  /*************************************************************************/
  /*************************************************************************/

  /* documentation is in ftsynth.h */

  FT_EXPORT_DEF( void )
  FT_GlyphSlot_Oblique( FT_GlyphSlot  slot )
  {
    FT_Matrix    transform;
    FT_Outline*  outline = &slot->outline;


    /* only oblique outline glyphs */
    if ( slot->format != FT_GLYPH_FORMAT_OUTLINE )
      return;

    /* we don't touch the advance width */

    /* For italic, simply apply a shear transform, with an angle */
    /* of about 12 degrees.                                      */

    transform.xx = 0x10000L;
    transform.yx = 0x00000L;

    transform.xy = 0x06000L;
    transform.yy = 0x10000L;

    FT_Outline_Transform( outline, &transform );
  }


  /*************************************************************************/
  /*************************************************************************/
  /****                                                                 ****/
  /****   EXPERIMENTAL EMBOLDENING/OUTLINING SUPPORT                    ****/
  /****                                                                 ****/
  /*************************************************************************/
  /*************************************************************************/


  /* documentation is in ftsynth.h */

  FT_EXPORT_DEF( void )
  FT_GlyphSlot_Embolden( FT_GlyphSlot  slot )
  {
    FT_Library  library = slot->library;
    FT_Face     face    = slot->face;
    FT_Error    error;
    FT_Pos      xstr, ystr;


    if ( slot->format != FT_GLYPH_FORMAT_OUTLINE &&
         slot->format != FT_GLYPH_FORMAT_BITMAP )
      return;

    /* some reasonable strength */
    xstr = FT_MulFix( face->units_per_EM,
                      face->size->metrics.y_scale ) / 24;
    ystr = xstr;

    if ( slot->format == FT_GLYPH_FORMAT_OUTLINE )
    {
      error = FT_Outline_Embolden( &slot->outline, xstr );
      /* ignore error */

      /* this is more than enough for most glyphs; if you need accurate */
      /* values, you have to call FT_Outline_Get_CBox                   */
      xstr = xstr * 2;
      ystr = xstr;
    }
    else if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
    {
      /* round to full pixels */
      xstr &= ~63;
      if ( xstr == 0 )
        xstr = 1 << 6;
      ystr &= ~63;

      error = FT_GlyphSlot_Own_Bitmap( slot );
      if ( error )
        return;

      error = FT_Bitmap_Embolden( library, &slot->bitmap, xstr, ystr );
      if ( error )
        return;
    }

    if ( slot->advance.x )
      slot->advance.x += xstr;

    if ( slot->advance.y )
      slot->advance.y += ystr;

    slot->metrics.width        += xstr;
    slot->metrics.height       += ystr;
    slot->metrics.horiBearingY += ystr;
    slot->metrics.horiAdvance  += xstr;
    slot->metrics.vertBearingX -= xstr / 2;
    slot->metrics.vertBearingY += ystr;
    slot->metrics.vertAdvance  += ystr;

    if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
      slot->bitmap_top += ystr >> 6;
  }

    static FT_Pos     Shadow_DispX = 4;   /* x-axis displacement of shadow effect */
    static FT_Pos     Shadow_DispY = 1;   /* y-axis displacement of shadow effect */
    static FT_Pos     Raise_DispX = 3;    /* x-axis displacement of raise and depress effect */
    static FT_Pos     Raise_DispY = 1;    /* x-axis displacement of raise and depress effect */
    static FT_Pos     Uniform_Thick = 1;  /* thickness of uniform effect */
    static FT_Pos     Outline_Thick = 1;  /* thickness of outline effect */  
#ifndef FE_DISABLE_EDGE_EFFECT_SUPPORT
	  FT_EXPORT_DEF( void )
	  FT_GlyphSlot_ShadowRight( FT_GlyphSlot  slot )
	  {
		FT_Library	library = slot->library;
		FT_Error	error	= FT_Err_Ok;
		FT_Pos		dx, dy;
  
		/* some reasonable strength */
		dx = Shadow_DispX;
		dy = Shadow_DispY;
  
		if ( slot->format == FT_GLYPH_FORMAT_OUTLINE )
		{
		  error = FT_Err_Unimplemented_Feature;
		  return;	/* unsupported */
		}
		else if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
		{
		  /* slot must be bitmap-owner */
		  if ( !( slot->internal->flags & FT_GLYPH_OWN_BITMAP ) )
		  {
			FT_Bitmap  bitmap;
  
			FT_Bitmap_New( &bitmap );
			error = FT_Bitmap_Copy( library, &slot->bitmap, &bitmap );
  
			if ( !error )
			{
			  slot->bitmap = bitmap;
			  slot->internal->flags |= FT_GLYPH_OWN_BITMAP;
			}
		  }
  
		  if ( !error )
			error = FT_Bitmap_Shadow( library, &slot->bitmap, dx, dy );
		}
		else
		  error = FT_Err_Invalid_Argument;
  
		/* modify the metrics accordingly */
		if ( !error )
		{
		  slot->advance.x += dx;
		  slot->advance.y += dy;
  
		  slot->metrics.width		 += dx;
		  slot->metrics.height		 += dy;
		  slot->metrics.horiBearingY += dy;
		  slot->metrics.horiAdvance  += dx;
		  slot->metrics.vertBearingX -= dx / 2;
		  slot->metrics.vertBearingY += dy;
		  slot->metrics.vertAdvance  += dy;
  
		  if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
			slot->bitmap_top += dy >> 6;
		}
	  }
  
  
	  FT_EXPORT_DEF( FT_Short )
	  FT_GlyphSlot_ShadowRight_Width_Inc( void )
	  {
		return (FT_Short) Shadow_DispX;
	  }
  
  
	  FT_EXPORT_DEF( void )
	  FT_GlyphSlot_ShadowLeft( FT_GlyphSlot  slot )
	  {
		FT_Library	library = slot->library;
		FT_Error	error	= FT_Err_Ok;
		FT_Pos		dx, dy;
  
		/* some reasonable strength */
		dx = Shadow_DispX;
		dy = Shadow_DispY;
  
		if ( slot->format == FT_GLYPH_FORMAT_OUTLINE )
		{
		  error = FT_Err_Unimplemented_Feature;
		  return;	/* unsupported */
		}
		else if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
		{
		  /* slot must be bitmap-owner */
		  if ( !( slot->internal->flags & FT_GLYPH_OWN_BITMAP ) )
		  {
			FT_Bitmap  bitmap;
  
			FT_Bitmap_New( &bitmap );
			error = FT_Bitmap_Copy( library, &slot->bitmap, &bitmap );
  
			if ( !error )
			{
			  slot->bitmap = bitmap;
			  slot->internal->flags |= FT_GLYPH_OWN_BITMAP;
			}
		  }
  
		  if ( !error )
			error = FT_Bitmap_Shadow( library, &slot->bitmap, -dx, dy );
		}
		else
		  error = FT_Err_Invalid_Argument;
  
		/* modify the metrics accordingly */
		if ( !error )
		{
		  slot->advance.x += dx;
		  slot->advance.y += dy;
  
		  slot->metrics.width		 += dx;
		  slot->metrics.height		 += dy;
		  slot->metrics.horiBearingY += dy;
		  slot->metrics.horiAdvance  += dx;
		  slot->metrics.vertBearingX -= dx / 2;
		  slot->metrics.vertBearingY += dy;
		  slot->metrics.vertAdvance  += dy;
  
		  if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
			slot->bitmap_top += dy >> 6;
		}
	  }
  
  
	  FT_EXPORT_DEF( FT_Short )
	  FT_GlyphSlot_ShadowLeft_Width_Inc( void )
	  {
		return (FT_Short) Shadow_DispX;
	  }
  
  
	  FT_EXPORT_DEF( void )
	  FT_GlyphSlot_Depressed( FT_GlyphSlot	slot )
	  {
		FT_Library	library = slot->library;
		FT_Error	error	= FT_Err_Ok;
		FT_Pos		dx, dy;
  
		/* some reasonable strength */
		dx = Raise_DispX;
		dy = Raise_DispY;
  
		if ( slot->format == FT_GLYPH_FORMAT_OUTLINE )
		{
		  error = FT_Err_Unimplemented_Feature;
		  return;	/* unsupported */
		}
		else if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
		{
		  /* slot must be bitmap-owner */
		  if ( !( slot->internal->flags & FT_GLYPH_OWN_BITMAP ) )
		  {
			FT_Bitmap  bitmap;
  
			FT_Bitmap_New( &bitmap );
			error = FT_Bitmap_Copy( library, &slot->bitmap, &bitmap );
  
			if ( !error )
			{
			  slot->bitmap = bitmap;
			  slot->internal->flags |= FT_GLYPH_OWN_BITMAP;
			}
		  }
  
		  if ( !error )
			error = FT_Bitmap_Raised( library, &slot->bitmap, -dx, -dy );
		}
		else
		  error = FT_Err_Invalid_Argument;
  
		/* modify the metrics accordingly */
		if ( !error )
		{
		  slot->advance.x += dx;
		  slot->advance.y += dy;
  
		  slot->metrics.width		 += dx;
		  slot->metrics.height		 += dy;
		  slot->metrics.horiBearingY += dy;
		  slot->metrics.horiAdvance  += dx;
		  slot->metrics.vertBearingX -= dx / 2;
		  slot->metrics.vertBearingY += dy;
		  slot->metrics.vertAdvance  += dy;
  
		  if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
			slot->bitmap_top += dy >> 6;
		}
	  }
  
  
	  FT_EXPORT_DEF( FT_Short )
	  FT_GlyphSlot_Depressed_Width_Inc( void )
	  {
		return (FT_Short) Raise_DispX;
	  }
  
  
	  FT_EXPORT_DEF( void )
	  FT_GlyphSlot_Raised( FT_GlyphSlot  slot )
	  {
		FT_Library	library = slot->library;
		FT_Error	error	= FT_Err_Ok;
		FT_Pos		dx, dy;
  
		/* some reasonable strength */
		dx = Raise_DispX;
		dy = Raise_DispY;
  
		if ( slot->format == FT_GLYPH_FORMAT_OUTLINE )
		{
		  error = FT_Err_Unimplemented_Feature;
		  return;	/* unsupported */
		}
		else if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
		{
		  /* slot must be bitmap-owner */
		  if ( !( slot->internal->flags & FT_GLYPH_OWN_BITMAP ) )
		  {
			FT_Bitmap  bitmap;
  
			FT_Bitmap_New( &bitmap );
			error = FT_Bitmap_Copy( library, &slot->bitmap, &bitmap );
  
			if ( !error )
			{
			  slot->bitmap = bitmap;
			  slot->internal->flags |= FT_GLYPH_OWN_BITMAP;
			}
		  }
  
		  if ( !error )
			error = FT_Bitmap_Raised( library, &slot->bitmap, dx, dy );
		}
		else
		  error = FT_Err_Invalid_Argument;
  
		/* modify the metrics accordingly */
		if ( !error )
		{
		  slot->advance.x += dx;
		  slot->advance.y += dy;
  
		  slot->metrics.width		 += dx;
		  slot->metrics.height		 += dy;
		  slot->metrics.horiBearingY += dy;
		  slot->metrics.horiAdvance  += dx;
		  slot->metrics.vertBearingX -= dx / 2;
		  slot->metrics.vertBearingY += dy;
		  slot->metrics.vertAdvance  += dy;
  
		  if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
			slot->bitmap_top += dy >> 6;
		}
	  }
  
  
	  FT_EXPORT_DEF( FT_Short )
	  FT_GlyphSlot_Raised_Width_Inc( void )
	  {
		return (FT_Short) Raise_DispX;
	  }
  
  
	  FT_EXPORT_DEF( void )
	  _ft_glyph_uniform(FT_GlyphSlot  slot, FT_Pos in_width)
	  {
		FT_Library	library = slot->library;
		FT_Error	error	= FT_Err_Ok;
		FT_Pos		width = in_width;
  
		if ( slot->format == FT_GLYPH_FORMAT_OUTLINE )
		{
		  error = FT_Err_Unimplemented_Feature;
		  return;	/* unsupported */
		}
		else if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
		{
		  /* slot must be bitmap-owner */
		  if ( !( slot->internal->flags & FT_GLYPH_OWN_BITMAP ) )
		  {
			FT_Bitmap  bitmap;
  
			FT_Bitmap_New( &bitmap );
			error = FT_Bitmap_Copy( library, &slot->bitmap, &bitmap );
  
			if ( !error )
			{
			  slot->bitmap = bitmap;
			  slot->internal->flags |= FT_GLYPH_OWN_BITMAP;
			}
		  }
  
		  if ( !error )
			error = FT_Bitmap_Uniform( library, &slot->bitmap, width );
		}
		else
		  error = FT_Err_Invalid_Argument;
  
		/* modify the metrics accordingly */
		if ( !error )
		{
		  slot->advance.x += 2 * width;
		  slot->advance.y += 2 * width;
  
		  slot->metrics.width		 += 2 * width;
		  slot->metrics.height		 += 2 * width;
		  slot->metrics.horiBearingY += 2 * width;
		  slot->metrics.horiAdvance  += 2 * width;
		  slot->metrics.vertBearingX -= width;
		  slot->metrics.vertBearingY += 2 * width;
		  slot->metrics.vertAdvance  += 2 * width;
  
		  if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
			slot->bitmap_top += width >> (6 - 1);
		}
	  }
  
  
	  FT_EXPORT_DEF( void )
	  FT_GlyphSlot_Uniform( FT_GlyphSlot  slot )
	  {
		_ft_glyph_uniform(slot, Uniform_Thick);
	  }
  
  
	  FT_EXPORT_DEF( void )
	  FT_GlyphSlot_Custom_Uniform( FT_GlyphSlot  slot , FT_Pos ui2_uniform_width)
	  {
		_ft_glyph_uniform(slot, (FT_Pos)ui2_uniform_width);
	  }
  
  
	  FT_EXPORT_DEF( FT_Short )
	  FT_GlyphSlot_Uniform_Width_Inc( void )
	  {
		return (FT_Short) 2 * Uniform_Thick;
	  }
  
  
	  FT_EXPORT_DEF( void )
	  _ft_glyph_outline( FT_GlyphSlot slot, FT_Pos in_width )
	  {
		FT_Library	library = slot->library;
		FT_Error	error	= FT_Err_Ok;
		FT_Pos		width = in_width;
  
		if ( slot->format == FT_GLYPH_FORMAT_OUTLINE )
		{
		  error = FT_Err_Unimplemented_Feature;
		  return;	/* unsupported */
		}
		else if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
		{
		  /* slot must be bitmap-owner */
		  if ( !( slot->internal->flags & FT_GLYPH_OWN_BITMAP ) )
		  {
			FT_Bitmap  bitmap;
  
			FT_Bitmap_New( &bitmap );
			error = FT_Bitmap_Copy( library, &slot->bitmap, &bitmap );
  
			if ( !error )
			{
			  slot->bitmap = bitmap;
			  slot->internal->flags |= FT_GLYPH_OWN_BITMAP;
			}
		  }
  
		  if ( !error )
			error = FT_Bitmap_Outline( library, &slot->bitmap, width );
		}
		else
		  error = FT_Err_Invalid_Argument;
  
		/* modify the metrics accordingly */
		if ( !error )
		{
		  slot->advance.x += 2 * width;
		  slot->advance.y += 2 * width;
  
		  slot->metrics.width		 += 2 * width;
		  slot->metrics.height		 += 2 * width;
		  slot->metrics.horiBearingY += 2 * width;
		  slot->metrics.horiAdvance  += 2 * width;
		  slot->metrics.vertBearingX -= width;
		  slot->metrics.vertBearingY += 2 * width;
		  slot->metrics.vertAdvance  += 2 * width;
  
		  if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
			slot->bitmap_top += width >> (6 - 1);
		}
	  }
  
	  FT_EXPORT_DEF( void )
	  FT_GlyphSlot_Outline( FT_GlyphSlot  slot )
	  {
		_ft_glyph_outline(slot, Outline_Thick);
	  }
  
	  FT_EXPORT_DEF( void )
	  FT_GlyphSlot_Custom_Outline( FT_GlyphSlot  slot, FT_Pos ui2_outline_width ) /*mod for bd_subtitle*/
	  {
		_ft_glyph_outline(slot, (FT_Pos)ui2_outline_width);
	  }
  
	  FT_EXPORT_DEF( FT_Short )
	  FT_GlyphSlot_Outline_Width_Inc( void )
	  {
		return (FT_Short) 2 * Outline_Thick;
	  }
#endif


/* END */
